#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Module contains CUPED-based data transformation methods for the experiment
acceleration.
"""
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd

from ambrosia import types
from ambrosia.tools.ab_abstract_component import AbstractVarianceReducer
from ambrosia.tools.back_tools import wrap_cols


class Cuped(AbstractVarianceReducer):
    """
    Class for data CUPED transformation.

    https://towardsdatascience.com/how-to-double-a-b-testing-speed-with-cuped-f80460825a90
    Y_hat = Y - theta * X
    theta := cov(X, Y) / Var(Y)
    It is important, that the mean covariance metric did not change over time!!!

    Parameters
    ----------
    verbose : bool, default: ``True``
        If ``True`` will print in sys.stdout the information
        about the variance reduction.

    Attributes
    ----------
    params : Dict
        Parameters of instance that will be updated after calling fit() method.
        Include:
        - target column name
        - covariate column name
        - name of column after the transformation
        - linear coefficient for CUPED transformation.
        - bias value for mean equality
    verbose : bool
        Verbose info flag.
    fitted : bool
        Flag if class was fitted.

    Examples
    --------
    Suppose we have the dataframe with users info which contains two columns:
    a "target" columns and a column with metric "income". Let us can assume,
    that over time, the average of the "income" values do not change. Then, we
    can use CUPED transformation based on "income" data to reduce "target"
    column variation.

    >>> cuped_transformer = Cuped(dataframe, 'target', verbose=True)
    >>> cuped_transformer.fit_transform(
    >>>     dataframe=dataframe
    >>>     target_column='target'
    >>>     covariate_column='income',
    >>>     transformed_name='cuped_target'
    >>>     inplace=True,
    >>> )

    Now in the dataframe a new column "cuped_target" appeared, we can use it
    to design our experiment and estimate variance reduction. For further CUPED
    usage in the future experiment, let us store the parameters:

    >>> cuped_transformer.store_params('cuped_transform_params.json')

    Now we conduct an experiment and want to transform our data to reduce its
    variation:

    >>> cuped_transformation = Cuped()
    >>> cuped_transformation.load_params('cuped_transform_params.json')
    >>> cuped_transformation.transform(
    >>>     dataframe=exp_results,
    >>>     inplace=True,
    >>> )

    Methods
    -------
    get_params_dict()
        Returns dictionary with params if fit() method has been previously
        called.
    load_params_dict(params)
        Load params from a dictionary.
    store_params(store_path)
        Store params to json file if fit() method has been previously called.
    load_params(load_path)
        Load params from a json file.
    fit(covariate_column)
        Fit model using a specific covariate column.
    transform(covariate_column, inplace, name)
        Transform target column after a class instance fitting.
    fit_transform(covariate_column, inplace, name)
        Combination of fit() and transform() methods.
    """

    THETA_NAME: str = "theta"
    BIAS_NAME: str = "bias"
    non_serializable_params: List = [THETA_NAME, BIAS_NAME]

    def __init__(self, verbose: bool = True) -> None:
        super().__init__(verbose)
        self.params["covariate_column"] = None
        self.params[Cuped.THETA_NAME] = None
        self.params[Cuped.BIAS_NAME] = None

    def __str__(self) -> str:
        return f"СUPED for {self.params['target_column']}"

    def __call__(self, y: np.ndarray, X: np.ndarray) -> np.ndarray:
        self._check_fitted()
        y_hat: np.ndarray = y - self.params[Cuped.THETA_NAME] * (X - self.params[Cuped.BIAS_NAME])
        return y_hat

    def get_params_dict(self) -> Dict:
        """
        Returns a dictionary with params.

        Returns
        -------
        params : Dict
            Dictionary with fitted params.
        """
        self._check_fitted()
        return {
            key: (value if key not in Cuped.non_serializable_params else value.tolist())
            for key, value in self.params.items()
        }

    def load_params_dict(self, params: Dict) -> None:
        """
        Load model parameters from the dictionary.

        Parameters
        ----------
        params : Dict
            Dictionary with params.
        """
        for parameter in self.params:
            if parameter in params:
                if parameter in Cuped.non_serializable_params:
                    self.params[parameter] = np.array(params[parameter])
                else:
                    self.params[parameter] = params[parameter]
            else:
                raise TypeError(f"params argument must contain: {parameter}")
        self.fitted = True

    def fit(
        self,
        dataframe: pd.DataFrame,
        target_column: types.ColumnNameType,
        covariate_column: types.ColumnNameType,
        transformed_name: Optional[types.ColumnNameType] = None,
    ) -> None:
        """
        Fit to calculate CUPED parameters for target column using given
        covariate column and data.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with data for the calculation of CUPED parameters.
        target_column : ColumnNameType
            Column from the dataframe, for which CUPED transformation will be
            applied.
        covariate_column : ColumnNameType
            Column which will be used as the covariate in CUPED transformation.
        transformed_name : ColumnNamesType, optional
            Name for the new transformed target column, if is not defined
            it will be generated automatically.
        """
        self._check_cols(dataframe, [target_column, covariate_column])
        covariance: pd.DataFrame = dataframe[[target_column, covariate_column]].cov()
        covariate_variance: float = covariance.loc[covariate_column, covariate_column]

        self.params[Cuped.THETA_NAME] = covariance.loc[target_column, covariate_column] / (
            super().EPSILON + covariate_variance
        )
        self.params[Cuped.BIAS_NAME] = np.mean(dataframe[covariate_column])
        self.params["target_column"] = target_column
        self.params["covariate_column"] = covariate_column
        self.params["transformed_name"] = transformed_name
        self.fitted = True

    def transform(
        self,
        dataframe: pd.DataFrame,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Make CUPED transformation for the target column.

        Could be performed inplace or not.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with data for CUPED transformation.
        inplace : bool, default: ``False``
            If is ``True``, then method returns ``None`` and
            sets a new column for the original dataframe.
            Otherwise return copied dataframe with a new column.
        """
        self._check_cols(dataframe, [self.params["target_column"], self.params["covariate_column"]])
        new_target: np.ndarray = self(
            dataframe[self.params["target_column"]], dataframe[self.params["covariate_column"]]
        )
        if self.verbose:
            old_variance: float = np.var(dataframe[self.params["target_column"]])
            new_variance: float = np.var(new_target)
            self._verbose(old_variance, new_variance)
        return self._return_result(dataframe, new_target, inplace)

    def fit_transform(
        self,
        dataframe,
        target_column,
        covariate_column: types.ColumnNameType,
        transformed_name: Optional[types.ColumnNameType] = None,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Combination of fit() and transform() methods.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with data for fitting and applying CUPED transformation.
        target_column : ColumnNameType
            Column from the dataframe, for which CUPED transformation will be
            applied.
        covariate_column : ColumnNameType
            Column which will be used as the covariate.
        transformed_name : ColumnNamesType, optional
            Name for the new transformed target column, if is not defined
            it will be generated automatically.
        inplace : bool, default: ``False``
            If is ``True``, then method returns ``None`` and
            sets a new column for the original dataframe.
            Otherwise return copied dataframe with a new column.
        """
        self.fit(dataframe, target_column, covariate_column, transformed_name)
        return self.transform(dataframe, inplace)


class MultiCuped(AbstractVarianceReducer):
    """
    Class for data Multi CUPED transformation.

    Y_hat = Y - X theta (Matrix multiplication)
    theta := argmin Var (Y - X theta)
    It is important, that the mean covariance metric do not change over time!!!


    Parameters
    ----------
    verbose : bool, default: ``True``
        If ``True`` will print in sys.stdout the information
        about the variance reduction.

    Attributes
    ----------
    params : Dict
        Parameters of instance that will be updated after calling fit() method.
        Include:
        - target column name
        - covariate columns names
        - name of column after the transformation
        - linear coefficients for Multi CUPED transformation.
        - bias value for mean equality
    verbose : bool
        Verbose info flag.
    fitted : bool
        Flag if class was fitted.

    Examples
    --------
    We have dataframe with users info with column 'target' and
    columns 'income' and 'age'. We can assume, that over time,
    the average of this covariate values does not change. Then, we can use
    multi cuped transformation to reduce variation.

    Suppose we have the dataframe with users info which contains two columns:
    a "target" columns and columns "income" and "age". Let us can assume,
    that over time, the average of the "income" and "age" values do not change.
    Then, we can use Multi CUPED transformation based on "income" and "age"
    data in order to reduce "target" column variation.

    >>> cuped_transformer = MultiCuped(verbose=True)
    >>> cuped_transformer.fit_transform(
    >>>     dataframe=dataframe
    >>>     target_column='target'
    >>>     ['income', 'age'],
    >>>     transformed_name='cuped_target'
    >>>     inplace=True,
    >>> )

    Now in the dataframe a new column "cuped_target" appeared, we can use it
    to design our experiment and estimate variance reduction. For further
    Multi CUPED usage in the future experiment, let us store the parameters:

    >>> cuped_transformer.store_params('cuped_transform_params.json')

    Now we conduct an experiment and want to transform our data to reduce its
    variation:

    >>> cuped_transformation = MultiCuped()
    >>> cuped_transformation.load_params('cuped_transform_params.json')
    >>> cuped_transformation.transform(
    >>>     exp_results,
    >>>     inplace=True,
    >>> )

    Methods
    -------
    get_params_dict()
        Returns dictionary with params if fit() method has been previously
        called.
    load_params_dict(params)
        Load params from a dictionary.
    store_params(store_path)
        Store params to json file if fit() method has been previously called.
    load_params(load_path)
        Load params from a json file.
    fit(covariate_column)
        Fit model using covariate columns.
    transform(covariate_column, inplace, name)
        Transform target column after a class instance fitting.
    fit_transform(covariate_column, inplace, name)
        Combination of fit() and transform() methods.
    """

    THETA_NAME: str = "theta"
    BIAS_NAME: str = "bias"
    non_serializable_params: List = [THETA_NAME, BIAS_NAME]

    def __init__(self, verbose: bool = True) -> None:
        super().__init__(verbose)
        self.params["covariate_columns"] = None
        self.params[MultiCuped.THETA_NAME] = None
        self.params[MultiCuped.BIAS_NAME] = None

    def __str__(self) -> str:
        return f"Multi СUPED for {self.params['target_column']}"

    def __call__(self, y: np.ndarray, X: np.ndarray) -> np.ndarray:
        self._check_fitted()
        y_hat: np.ndarray = y - (X @ self.params[MultiCuped.THETA_NAME]).reshape(-1) + self.params[MultiCuped.BIAS_NAME]
        return y_hat

    def get_params_dict(self) -> Dict:
        """
        Returns a dictionary with params.

        Returns
        -------
        params : Dict
            Dictionary with fitted params.
        """
        self._check_fitted()
        return {
            key: (value if key not in MultiCuped.non_serializable_params else value.tolist())
            for key, value in self.params.items()
        }

    def load_params_dict(self, params: Dict) -> None:
        """
        Load model parameters from the dictionary.

        Parameters
        ----------
        params : Dict
            Dictionary with params.
        """
        for parameter in self.params:
            if parameter in params:
                if parameter in MultiCuped.non_serializable_params:
                    self.params[parameter] = np.array(params[parameter])
                else:
                    self.params[parameter] = params[parameter]
            else:
                raise TypeError(f"params argument must contain: {parameter}")
        self.fitted = True

    def fit(
        self,
        dataframe: pd.DataFrame,
        target_column: types.ColumnNameType,
        covariate_columns: types.ColumnNamesType,
        transformed_name: Optional[types.ColumnNameType] = None,
    ) -> None:
        """
        Fit to calculate Multi CUPED parameters for target column using selected
        covariate columns.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with data for the calculation of CUPED parameters.
        target_column : ColumnNameType
            Column from the dataframe, for which CUPED transformation will be
            applied.
        covariate_columns : ColumnNamesType
            Columns which will be used as the covariates in Multi CUPED
            transformation.
        transformed_name : ColumnNamesType, optional
            Name for the new transformed target column, if is not defined
            it will be generated automatically.
        """
        covariate_columns = wrap_cols(covariate_columns)
        cols_concat: List = [target_column] + covariate_columns
        self._check_cols(dataframe, cols_concat)
        covariance: np.ndarray = dataframe[cols_concat].cov()
        matrix: np.ndarray = covariance.loc[covariate_columns, covariate_columns]
        num_features: int = len(covariate_columns)
        covariance_target: np.ndarray = covariance.loc[covariate_columns, target_column].values.reshape(
            num_features, -1
        )

        self.params[MultiCuped.THETA_NAME] = np.linalg.inv(matrix) @ covariance_target
        self.params[MultiCuped.BIAS_NAME]: np.ndarray = (
            (dataframe[covariate_columns].values @ self.params[MultiCuped.THETA_NAME]).reshape(-1).mean()
        )
        self.params["target_column"] = target_column
        self.params["covariate_columns"] = covariate_columns
        self.params["transformed_name"] = transformed_name
        self.fitted = True

    def transform(
        self,
        dataframe: pd.DataFrame,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Make Multi CUPED transformation for the target column.

        Could be performed inplace or not.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with data for Multi CUPED transformation.
        inplace : bool, default: ``False``
            If is ``True``, then method returns ``None`` and
            sets a new column for the original dataframe.
            Otherwise return copied dataframe with a new column.
        """
        self._check_cols(dataframe, [self.params["target_column"]] + self.params["covariate_columns"])
        self._check_fitted()
        new_target: np.ndarray = self(
            dataframe[self.params["target_column"]].values, dataframe[self.params["covariate_columns"]].values
        )
        if self.verbose:
            old_variance: float = np.var(dataframe[self.params["target_column"]])
            new_variance: float = np.var(new_target)
            self._verbose(old_variance, new_variance)
        return self._return_result(dataframe, new_target, inplace)

    def fit_transform(
        self,
        dataframe: pd.DataFrame,
        target_column: types.ColumnNameType,
        covariate_columns: types.ColumnNamesType,
        transformed_name: Optional[types.ColumnNameType] = None,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Combination of fit() and transform() methods.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with data for fitting and applying Multi CUPED transformation.
        target_column : ColumnNameType
            Column from the dataframe, for which CUPED transformation will be
            applied.
        covariate_column : ColumnNameType
            Column which will be used as the covariate.
        transformed_name : ColumnNamesType, optional
            Name for the new transformed target column, if is not defined
            it will be generated automatically.
        inplace : bool, default: ``False``
            If is ``True``, then method returns ``None`` and
            sets a new column for the original dataframe.
            Otherwise return copied dataframe with a new column.
        """
        self.fit(dataframe, target_column, covariate_columns, transformed_name)
        return self.transform(dataframe, inplace)



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Iterable, Tuple

import pandas as pd
import scipy.stats as sps

import ambrosia.tools.theoretical_tools as theory_pkg
from ambrosia import types
from ambrosia.tools.import_tools import spark_installed

if spark_installed():
    import pyspark.sql.functions as funcs


def get_stats_from_table(dataframe: types.SparkDataFrame, column: types.ColumnNameType) -> Tuple[float, float]:
    """
    Get table for designing samples size for experiment.
    """
    stats = dataframe.select(
        funcs.mean(funcs.col(column)).alias("mean"), funcs.stddev(funcs.col(column)).alias("std")
    ).collect()
    mean = stats[0]["mean"]
    std = stats[0]["std"]
    return mean, std


def design_groups_size(
    dataframe: types.SparkDataFrame,
    column: types.ColumnNameType,
    effects: Iterable[float],
    second_errors: Iterable[float],
    first_errors: Iterable[float] = (0.05,),
) -> pd.DataFrame:
    """
    Create pandas dataframe with designed groups sizes for metric based on data
    in spark dataframe.

    Results in returned dataframe depend on desired effects, first and second type errors,
    and statistic of metric from given spark table.

    Parameters
    ----------
    dataframe : types.SparkDataFrame
        Table for designing experiment
    column : types.ColumnNameType
        Column, containg metric for designing
    effects : Iterable[float]
        List of effects which we want to catch.
        e.x.: [1.01, 1.02, 1.05]
    second_errors : Iterable[float]
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    first_errors : Iterable[float], default: ``(0.05,)``
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]

    Returns
    -------
    df_results : pd.DataFrame
        Table with minimal sample sizes for each effect and errors pair based on input data.
    """
    mean, std = get_stats_from_table(dataframe, column)
    return theory_pkg.get_table_sample_size(mean, std, effects, first_errors, second_errors)


def design_effect(
    dataframe: types.SparkDataFrame,
    column: types.ColumnNameType,
    sample_sizes: Iterable[int],
    second_errors: Iterable[float],
    first_errors: Iterable[float] = (0.05,),
) -> pd.DataFrame:
    """
    Create pandas dataframe with designed effects for metric based on data in spark dataframe.
    Results in returned dataframe depend on sample sizes, first and second type errors,
    and statistic of metric from given spark table.

    Parameters
    ----------
    dataframe : types.SparkDataFrame
        Spark table with data for experiment design
    column : types.ColumnNameType
        Column, containg metric for designing
    sample_sizes : Iterable[int]
        List of sample sizes which we want to check.
        e.x.: [100, 200, 1000]
    second_errors : Iterable[float]
        2nd type errors.
        e.x.: [0.01, 0.05, 0.1]
    first_errors : Iterable[float], default: ``(0.05,)``
        1st type errors.
        e.x.: [0.01, 0.05, 0.1]

    Returns
    -------
    df_results : pd.DataFrame
        Table with minimal effects for each sample size and errors pair based on input data.
    """
    mean, std = get_stats_from_table(dataframe, column)
    return theory_pkg.get_minimal_effects_table(mean, std, sample_sizes, first_errors, second_errors)


def design_power(
    dataframe: types.SparkDataFrame,
    column: types.ColumnNameType,
    sample_sizes: Iterable[int],
    effects: Iterable[float],
    first_errors: Iterable[float] = (0.05,),
) -> pd.DataFrame:
    """
    Create pandas dataframe with designed power for metric based on data in spark dataframe.
    Results in returned dataframe depend on sample sizes, disered effects, first type errors,
    and statistic of metric from given spark table.

    Parameters
    ----------
    dataframe : types.SparkDataFrame
        Spark table with data for experiment design
    column : types.ColumnNameType
        Column, containg metric for designing
    sample_sizes : Iterable[int]
        List of sample sizes which we want to check.
        e.x.: [100, 200, 1000]
    effects : Iterable[float]
        Iterable object with
        e.x.: [1.01, 1.02, 1.05]
    first_errors : Iterable[float], default: ``(0.05,)``
        1st and 2nd type errors.
        e.x.: [0.01, 0.05, 0.1]

    Returns
    -------
    df_results : pd.DataFrame
        Table with power for each sample size, effect and first type error based on input data.
    """
    mean, std = get_stats_from_table(dataframe, column)
    return theory_pkg.get_power_table(mean, std, sample_sizes, effects, first_errors)


def ttest_spark(
    first_group: types.SparkDataFrame, second_group: types.SparkDataFrame, column: types.ColumnNameType
) -> Tuple[float, float]:
    """
    T-test for independent groups.

    Parameters
    ----------
    first_group : Spark Data Frame
        Data Frame for first group
    first_group : Spark Data Frame
        Data Frame for second group
    column : Column Type
        Column to be tested

    Returns
    -------
    statistic, pvalue : Tuple[float, float]
        T-test result
    """
    mean_1, std_1 = get_stats_from_table(first_group, column)
    mean_2, std_2 = get_stats_from_table(second_group, column)
    n_obs_1: int = first_group.count()
    n_obs_2: int = second_group.count()
    return sps.ttest_ind_from_stats(mean_1, std_1, n_obs_1, mean_2, std_2, n_obs_2)



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Module contains `Preprocessor` class that combines all data preprocessing
methods in one single chain pipeline. The resulting pipeline allows one to
consistently apply the desired transformations to the data, including outliers
removal, data aggregation and target metric transformations for the variance
reduction.
"""
from __future__ import annotations

import inspect
import json
import sys
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd

from ambrosia import types
from ambrosia.preprocessing.aggregate import AggregatePreprocessor
from ambrosia.preprocessing.cuped import Cuped, MultiCuped
from ambrosia.preprocessing.robust import IQRPreprocessor, RobustPreprocessor
from ambrosia.preprocessing.transformers import BoxCoxTransformer, LinearizationTransformer, LogTransformer


class Preprocessor:
    """
    Preprocessor class, implementation is based on the chain pattern.

    Parameters
    ----------
    dataframe : pd.DataFrame
        Table with data used for further transformations.
    verbose : bool, default: ``True``
        If ``True`` will print in sys.stdout the information
        about the variance reduction.

    Attributes
    ----------
    dataframe : pd.DataFrame
        Table with data for transformations.
    transformers : List of transformations
        List of transformation that have been called before.
    verbose : bool
        Verbose info flag.

    Examples
    --------
    >>> transformer = Preprocessor(dataframe)
    >>> transformer.aggregate(aggregate_params)
    >>>            .robust(robust_params)
    >>>            .cuped(cuped_params)
    >>>            .data()

    Methods
    -------
    data(copy=True)
        Returns a copy or a link for the stored dataframe.
    aggregate(groupby_columns, categorial_method, real_method, agg_params,
              real_cols, categorial_cols)
        Aggreagate data by columns.
    robust(column_names, alpha=0.05)
        Make a robust preprocessing of data.
    iqr(column_names, alpha=0.05)
        Make an IQR preprocessing of data.
    boxcox(column_names, alpha=0.05)
        Make a Box-Cox transformation.
    log(column_names, alpha=0.05)
        Make a log transformation.
    cuped(target, by, name, load_path)
        Make CUPED transformation for the stored dataframe.
    multicuped(target, by, name, load_path)
        Make Multi CUPED transformation for the stored dataframe.
    transformations()
        Returns a list of transformations.
    store_transformations(store_path)
        Store transformations in a json file.
    load_transformations(load_path)
        Load transformations from a json file.
    apply_transformations()
        Apply transformations for the stored dataframe.
    transform_from_config(load_path)
        Transform inner data frame using pre-saved config file.
    """

    def __len__(self) -> int:
        return len(self.dataframe)

    def __init__(self, dataframe: pd.DataFrame, verbose: bool = True) -> None:
        self.dataframe = dataframe.copy()
        self.transformers = []
        self.verbose = verbose

    def data(self, copy: bool = True):
        """
        Return the inner data frame.

        Use after all transformations to get transformed data.

        Parameters
        ----------
        copy : bool, default: ``True``
            If true returns copy, otherwise link

        Returns
        -------
        dataframe : pd.DataFrame
            Table with the modified data after the sequential preprocessing.
        """
        return self.dataframe.copy() if copy else self.dataframe

    def aggregate(
        self,
        groupby_columns: Optional[types.ColumnNamesType] = None,
        categorial_method: types.MethodType = "mode",
        real_method: types.MethodType = "sum",
        agg_params: Optional[Dict] = None,
        real_cols: Optional[types.ColumnNamesType] = None,
        categorial_cols: Optional[types.ColumnNamesType] = None,
        load_path: Optional[Path] = None,
    ) -> Preprocessor:
        """
        Make an aggregation of the dataframe.

        Parameters
        ----------
        groupby_columns : List of columns, optional
            Columns for GROUP BY.
        categorial_method : types.MethodType, default: ``"mode"``
            Aggregation method  that will be applied for all selected
            categorial variables.
        real_method : types.MethodType, default: ``"sum"``
            Aggregation method  that will be applied for all selected
            real variables.
        agg_params : Dict, optional
            Dictionary with aggregation parameters.
        real_cols : types.ColumnNamesType, optional
            Columns with real metrics.
            Overriden by ``agg_params`` parameter and could be passed if
            expected default aggregation behavior.
        categorial_cols : types.ColumnNamesType, optional
            Columns with categorial metrics
            Overriden by ``agg_params`` parameter and could be passed if
            expected default aggregation behavior.

        Returns
        -------
        self : Preprocessor
            Instance object
        """
        transformer = AggregatePreprocessor(categorial_method, real_method)
        if load_path is None:
            self.dataframe = transformer.fit_transform(
                self.dataframe, groupby_columns, agg_params, real_cols, categorial_cols
            )
        else:
            transformer.load_params(load_path)
            self.dataframe = transformer.transform(self.dataframe)
        self.transformers.append(transformer)
        return self

    def robust(
        self,
        column_names: Optional[types.ColumnNamesType] = None,
        alpha: Union[float, np.ndarray] = 0.05,
        tail: str = "both",
        load_path: Optional[Path] = None,
    ) -> Preprocessor:
        """
        Make a robust preprocessing of the selected columns to remove outliers.

        Removes objects from the dataframe which are in the head, end or
        both tail parts of the selected metrics distributions.

        Parameters
        ----------
        column_names : ColumnNamesType
            One or number of columns in the dataframe.
        alpha : Union[float, np.ndarray], default: ``0.05``
            The percentage of removed data from head and tail.
        tail : str, default: ``"both"``
            Part of distribution to be removed.
            Can be ``"left"``, ``"right"`` or ``"both"``.
        load_path : Path, optional
            Path to json file with parameters.

        Returns
        -------
        self : Preprocessor
            Instance object
        """
        transformer = RobustPreprocessor(verbose=self.verbose)
        if load_path is None:
            transformer.fit_transform(self.dataframe, column_names, alpha, tail, inplace=True)
        else:
            transformer.load_params(load_path)
            transformer.transform(self.dataframe, inplace=True)
        self.transformers.append(transformer)
        return self

    def iqr(
        self,
        column_names: Optional[types.ColumnNamesType] = None,
        load_path: Optional[Path] = None,
    ) -> Preprocessor:
        """
        Make an IQR preprocessing of the selected columns to remove outliers.

        Removes objects from the dataframe which are behind boxplot maximum
        and minimum of the selected metrics distributions.

        Parameters
        ----------
        column_names : ColumnNamesType, optional
            One or number of columns in the dataframe.
        load_path : Path, optional
            Path to json file with parameters.

        Returns
        -------
        self : Preprocessor
            Instance object
        """
        transformer = IQRPreprocessor(verbose=self.verbose)
        if load_path is None:
            transformer.fit_transform(self.dataframe, column_names, inplace=True)
        else:
            transformer.load_params(load_path)
            transformer.transform(self.dataframe, inplace=True)
        self.transformers.append(transformer)
        return self

    def boxcox(
        self,
        column_names: Optional[types.ColumnNamesType] = None,
        load_path: Optional[Path] = None,
    ) -> Preprocessor:
        """
        Make a Box-Cox transformation on the selected columns.

        Optimal transformation parameters are selected automatically.

        Parameters
        ----------
        column_names : ColumnNamesType, optional
            One or number of columns in the dataframe.
        load_path : Path, optional
            Path to json file with parameters.

        Returns
        -------
        self : Preprocessor
            Instance object
        """
        transformer = BoxCoxTransformer()
        if load_path is None:
            transformer.fit_transform(self.dataframe, column_names, inplace=True)
        else:
            transformer.load_params(load_path)
            transformer.transform(self.dataframe, inplace=True)
        self.transformers.append(transformer)
        return self

    def log(
        self,
        column_names: Optional[types.ColumnNamesType] = None,
        load_path: Optional[Path] = None,
    ) -> Preprocessor:
        """
        Make a logarithmic transformation on the selected columns.

        Parameters
        ----------
        column_names : ColumnNamesType, optional
            One or number of columns in the dataframe.
        load_path : Path, optional
            Path to json file with parameters.

        Returns
        -------
        self : Preprocessor
            Instance object
        """
        transformer = LogTransformer()
        if load_path is None:
            transformer.fit_transform(self.dataframe, column_names, inplace=True)
        else:
            transformer.load_params(load_path)
            transformer.transform(self.dataframe, inplace=True)
        self.transformers.append(transformer)
        return self

    def cuped(
        self,
        target: Optional[types.ColumnNameType] = None,
        by: Optional[types.ColumnNameType] = None,
        transformed_name: Optional[types.ColumnNameType] = None,
        load_path: Optional[Path] = None,
    ) -> Preprocessor:
        """
        Make CUPED transformation on the selected column.

        Parameters
        ----------
        target : ColumnNameType
            Column from the dataframe, for which CUPED transformation will be
            applied.
        by : ColumnNameType
            Covariance column in the dataframe.
        transformed_name : types.ColumnNameType, optional
            Name for the new transformed target column, if is not defined
            it will be generated automatically.
        load_path : Path, optional
            Path to json file with parameters.

        Returns
        -------
        self : Preprocessor
            Instance object
        """
        transformer = Cuped(verbose=self.verbose)
        if load_path is None:
            transformer.fit_transform(self.dataframe, target, by, transformed_name, inplace=True)
        else:
            transformer.load_params(load_path)
            transformer.transform(self.dataframe, inplace=True)
        self.transformers.append(transformer)
        return self

    def multicuped(
        self,
        target: Optional[types.ColumnNameType] = None,
        by: Optional[types.ColumnNamesType] = None,
        transformed_name: Optional[types.ColumnNameType] = None,
        load_path: Optional[Path] = None,
    ) -> Preprocessor:
        """
        Make Multi CUPED transformation on the selected column.

        Parameters
        ----------
        target : ColumnNameType
            Column from the dataframe, for which CUPED transformation will be
            applied.
        by : ColumnNameType
            Covariance columns in the dataframe.
        transformed_name : types.ColumnNameType, optional
            Name for the new transformed target column, if is not defined
            it will be generated automatically.
        load_path : Path, optional
            Path to json file with parameters.

        Returns
        -------
        self : Preprocessor
            Instance object
        """
        transformer = MultiCuped(verbose=self.verbose)
        if load_path is None:
            transformer.fit_transform(self.dataframe, target, by, transformed_name, inplace=True)
        else:
            transformer.load_params(load_path)
            transformer.transform(self.dataframe, inplace=True)
        self.transformers.append(transformer)
        return self

    def linearize(
        self,
        numerator: types.ColumnNameType,
        denominator: types.ColumnNameType,
        transformed_name: Optional[types.ColumnNameType] = None,
        load_path: Optional[Path] = None,
    ) -> Preprocessor:
        """
        Linearize a ratio metric for use in A/B testing.

        Computes a per-unit linearized value that is approximately normally
        distributed, enabling correct t-test usage for ratio metrics:

            linearized_i = numerator_i - ratio * denominator_i

        where ratio = mean(numerator) / mean(denominator) is estimated on
        the data passed to this ``Preprocessor`` instance (reference / control data).

        Parameters
        ----------
        numerator : ColumnNameType
            Column name of the ratio numerator (e.g. ``"revenue"``).
        denominator : ColumnNameType
            Column name of the ratio denominator (e.g. ``"orders"``).
        transformed_name : ColumnNameType, optional
            Name for the new linearized column. Defaults to
            ``"{numerator}_lin"``.
        load_path : Path, optional
            Path to a json file with pre-fitted parameters.

        Returns
        -------
        self : Preprocessor
            Instance object.
        """
        transformer = LinearizationTransformer()
        if load_path is None:
            transformer.fit_transform(self.dataframe, numerator, denominator, transformed_name, inplace=True)
        else:
            transformer.load_params(load_path)
            transformer.transform(self.dataframe, inplace=True)
        self.transformers.append(transformer)
        return self

    def transformations(self) -> List:
        """
        List of all transformations which were called.

        Returns
        -------
        transformers : List[object]
            List of executed transformations
        """
        return self.transformers

    def store_transformations(self, store_path: Path) -> None:
        """
        Store transformations with parameters in the json file.

        Parameters
        ----------
        store_path : Path
            Path to a json file where transformations will be stored
        """
        if len(self.transformers) == 0:
            raise ValueError("No transformations have been made yet.")
        transformations_counter = {}
        transformations_config = {}
        for transformer in self.transformers:
            alias = transformer.__class__.__name__
            if alias in transformations_counter:
                transformations_counter[alias] += 1
            else:
                transformations_counter[alias] = 1
            alias += "_" + str(transformations_counter[alias])
            transformations_config[alias] = transformer.get_params_dict()

        with open(store_path, "w+") as file:
            json.dump(transformations_config, file)

    def load_transformations(self, load_path: Path) -> None:
        """
        Load pre-saved transformations from the json file.

        Parameters
        ----------
        load_path : Path
            Path to a json file where transformations are stored
        """
        with open(load_path, "r+") as file:
            params = json.load(file)
        for key, value in params.items():
            class_alias = "".join(filter(str.isalpha, key))
            transformer = getattr(sys.modules[__name__], class_alias)
            kwargs = {}
            if "verbose" in inspect.signature(transformer).parameters:
                kwargs["verbose"] = self.verbose
            transformer = transformer(**kwargs)
            transformer.load_params_dict(value)
            self.transformers.append(transformer)

    def apply_transformations(self) -> pd.DataFrame:
        """
        Apply all transformations to the inner data frame.

        Returns
        -------
        dataframe : pd.DataFrame
            Transformed inner data frame
        """
        for transformer in self.transformers:
            if isinstance(transformer, AggregatePreprocessor):
                self.dataframe = transformer.transform(self.dataframe)
            else:
                transformer.transform(self.dataframe, inplace=True)
        return self.data()

    def transform_from_config(self, load_path: Path) -> pd.DataFrame:
        """
        Run transformations from the config file on the internal data frame.

        Parameters
        ----------
        load_path : Path
            Path to a json file where transformations are stored.

        Returns
        -------
        dataframe : pd.DataFrame
            Transformed inner data frame
        """
        self.load_transformations(load_path)
        return self.apply_transformations()



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Module contains tools for outliers removal from data during a
preprocessing task.
"""
from typing import Dict, Iterable, List, Union

import numpy as np
import pandas as pd

from ambrosia import types
from ambrosia.tools import log
from ambrosia.tools.ab_abstract_component import AbstractFittableTransformer
from ambrosia.tools.back_tools import wrap_cols


class RobustPreprocessor(AbstractFittableTransformer):
    """
    Unit for simple robust transformation for avoiding outliers in data.

    It cuts the alpha percentage of distribution from head, tail or both sides
    for each given metric.
    The data distribution structure assumed to present as small alpha
    part of outliers, followed by the normal part of the data with another
    alpha part of outliers at the end of the distribution.

    Parameters
    ----------
    verbose : bool, default: ``True``
        If ``True`` will show info about the transformation of passed columns.

    Attributes
    ----------
    params : Dict
        Dictionary with operational parameters of the instance.
        Updated after calling the ``fit`` method.
    verbose : bool
        Verbose info flag.
    available_tails : List
        List of the available tail type names to preprocess
    non_serializable_params: List
        List of the class parameters that should be converted to lists
        in order to serialize.
    fitted : bool
        Fit flag.

    Examples
    --------
    >>> robust = RobustPreprocessor(verbose=True)
    >>> robust.fit(dataframe, ['column1', 'column2'], alpha=0.05)
    >>> robust.transform(dataframe, inplace=True)

    You can pass one or number of columns, if several columns are passed
    it will drop in total alpha percent of extreme values for each column.
    """

    available_tails: List = ["both", "left", "right"]
    non_serializable_params: List = ["alpha", "quantiles"]

    def __str__(self) -> str:
        return "Robust preprocessing"

    def __init__(self, verbose: bool = True) -> None:
        """
        RobustPreprocessor class constructor.
        """
        self.params = {
            "tail": None,
            "column_names": None,
            "alpha": None,
            "quantiles": None,
        }
        self.verbose = verbose
        super().__init__()

    def get_params_dict(self) -> Dict:
        """
        Returns a dictionary with params.

        Returns
        -------
        params : Dict
            Dictionary with fitted params.
        """
        self._check_fitted()
        return {
            key: (value if key not in RobustPreprocessor.non_serializable_params else value.tolist())
            for key, value in self.params.items()
        }

    def load_params_dict(self, params: Dict) -> None:
        """
        Load prefitted parameters form a dictionary.

        Parameters
        ----------
        params : Dict
            Dictionary with prefitted params.
        """
        for parameter in self.params:
            if parameter in params:
                if parameter in RobustPreprocessor.non_serializable_params:
                    self.params[parameter] = np.array(params[parameter])
                else:
                    self.params[parameter] = params[parameter]
            else:
                raise TypeError(f"params argument must contain: {parameter}")
        self.fitted = True

    def __wrap_alpha(self, alpha: Union[float, Iterable]) -> np.ndarray:
        columns_num = len(self.params["column_names"])
        if isinstance(alpha, float):
            alpha = np.array([alpha] * columns_num)
        elif isinstance(alpha, Iterable):
            alpha = np.array(alpha)
        else:
            raise ValueError("Alpha parameter must be float or an iterable")
        if len(alpha) != columns_num:
            raise ValueError("Alpha length must be equal to the columns number")
        if (alpha < 0).any() or (alpha >= 0.5).any():
            raise ValueError(f"Alpha value must be from 0 to 0.5, but alpha vector = {alpha}")
        return alpha

    def __check_tail(self, tail: str) -> str:
        if tail not in self.available_tails:
            raise ValueError(f"tail must be one of {RobustPreprocessor.available_tails}")
        return tail

    def __calculate_quantiles(
        self,
        dataframe: pd.DataFrame,
    ) -> None:
        columns_num = len(self.params["column_names"])
        if self.params["tail"] == "both":
            self.params["quantiles"] = np.zeros((columns_num, 2))
            for num, col in enumerate(self.params["column_names"]):
                alpha = self.params["alpha"][num] / 2
                self.params["quantiles"][num, :] = np.quantile(dataframe[col].values, [alpha, 1 - alpha])
        else:
            self.params["quantiles"] = np.zeros((columns_num, 1))
            for num, col in enumerate(self.params["column_names"]):
                alpha = self.params["alpha"][num] if self.params["tail"] == "left" else 1 - self.params["alpha"][num]
                self.params["quantiles"][num] = np.quantile(dataframe[col].values, alpha)

    def fit(
        self,
        dataframe: pd.DataFrame,
        column_names: types.ColumnNamesType,
        alpha: Union[float, np.ndarray] = 0.05,
        tail: str = "both",
    ):
        """
        Fit to calculate robust parameters for the selected columns.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to calculate quantiles.
        column_names : ColumnNamesType
            One or number of columns in the dataframe.
        alpha : Union[float, np.ndarray], default: ``0.05``
            The percentage of removed data from head and tail.
        tail : str, default: ``"both"``
            Part of distribution to be removed.
            Can be ``"left"``, ``"right"`` or ``"both"``.

        Returns
        -------
        self : object
            Instance object.
        """
        self.params["column_names"] = wrap_cols(column_names)
        self._check_cols(dataframe, self.params["column_names"])
        self.params["alpha"] = self.__wrap_alpha(alpha)
        self.params["tail"] = self.__check_tail(tail)
        self.__calculate_quantiles(dataframe)
        self.fitted = True
        return self

    def transform(self, dataframe: pd.DataFrame, inplace: bool = False) -> Union[pd.DataFrame, None]:
        """
        Remove objects from the dataframe which are in the head, tail or both
        alpha parts of chosen metrics distributions.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to transform.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self._check_fitted()
        self._check_cols(dataframe, self.params["column_names"])
        if self.verbose:
            prev_stats: List[Dict[str, float]] = log.RobustLogger.get_stats(dataframe, self.params["column_names"])

        transformed: pd.DataFrame = dataframe if inplace else dataframe.copy()
        if self.params["tail"] == "both":
            mask: pd.Series = (transformed[self.params["column_names"]] < self.params["quantiles"][:, 0]).any(
                axis=1
            ) | (transformed[self.params["column_names"]] > self.params["quantiles"][:, 1]).any(axis=1)
        elif self.params["tail"] == "left":
            mask = (transformed[self.params["column_names"]] < self.params["quantiles"].T).any(axis=1)
        elif self.params["tail"] == "right":
            mask = (transformed[self.params["column_names"]] > self.params["quantiles"].T).any(axis=1)
        bad_ids = transformed.loc[mask].index
        transformed.drop(bad_ids, inplace=True)

        if self.verbose:
            log.info_log(
                f"""Making {self.params['tail']}-tail robust transformation of columns {self.params['column_names']}
                 with alphas = {np.round(self.params['alpha'], 3)}"""
            )
            new_stats: Dict[str, float] = log.RobustLogger.get_stats(transformed, self.params["column_names"])
            log.RobustLogger.verbose_list(prev_stats, new_stats, self.params["column_names"])
        return None if inplace else transformed

    def fit_transform(
        self,
        dataframe: pd.DataFrame,
        column_names: types.ColumnNamesType,
        alpha: Union[float, np.ndarray] = 0.05,
        tail: str = "both",
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Fit preprocessor parameters using given dataframe and transform it.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to calculate quantiles and for further transformation.
        column_names : ColumnNamesType
            One or number of columns in the dataframe.
        alpha : Union[float, np.ndarray], default: ``0.05``
            The percentage of removed data from head and tail.
        tail : str, default: ``"both"``
            Part of distribution to be removed.
            Can be ``"left"``, ``"right"`` or ``"both"``.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self.fit(dataframe, column_names, alpha, tail)
        return self.transform(dataframe, inplace)


class IQRPreprocessor(AbstractFittableTransformer):
    """
    Unit for IQR transformation of the data to exclude outliers.

    It cuts the points from the distribution which are behind the range of
    0.25 quantile - 1,5 * iqr and 0.75 quantile + 1,5 * iqr
    for each given metric.


    Parameters
    ----------
    verbose : bool, default: ``True``
        If ``True`` will show info about the transformation of passed columns.

    Attributes
    ----------
    params : Dict
        Dictionary with operational parameters of the instance.
        Updated after calling the ``fit`` method.
    verbose : bool
        Verbose info flag.
    non_serializable_params: List
        List of the class parameters that should be converted to lists
        in order to serialize.
    fitted : bool
        Fit flag.

    Examples
    --------
    >>> iqr = IQRPreprocessor(verbose=True)
    >>> iqr.fit(dataframe, ['column1', 'column2'])
    >>> iqr.transform(dataframe, inplace=True)

    You can pass one or number of columns, if several columns are passed
    it will drop extreme values for each column.
    """

    non_serializable_params: List = ["medians", "quartiles"]

    def __str__(self) -> str:
        return "IQR outliers preprocessing"

    def __init__(self, verbose: bool = True) -> None:
        """
        IQRPreprocessor class constructor.
        """
        self.params = {"column_names": None, "medians": None, "quartiles": None}
        self.verbose = verbose
        super().__init__()

    def get_params_dict(self) -> Dict:
        """
        Returns a dictionary with params.

        Returns
        -------
        params : Dict
            Dictionary with fitted params.
        """
        self._check_fitted()
        return {
            key: (value if key not in IQRPreprocessor.non_serializable_params else value.tolist())
            for key, value in self.params.items()
        }

    def load_params_dict(self, params: Dict) -> None:
        """
        Load prefitted parameters form a dictionary.

        Parameters
        ----------
        params : Dict
            Dictionary with prefitted params.
        """
        for parameter in self.params:
            if parameter in params:
                if parameter in IQRPreprocessor.non_serializable_params:
                    self.params[parameter] = np.array(params[parameter])
                else:
                    self.params[parameter] = params[parameter]
            else:
                raise TypeError(f"params argument must contain: {parameter}")
        self.fitted = True

    def __calculate_params(
        self,
        dataframe: pd.DataFrame,
    ):
        X: np.ndarray = dataframe[self.params["column_names"]].values
        self.params["quartiles"] = np.quantile(X, (0.25, 0.75), axis=0).T
        self.params["medians"] = np.median(X, axis=0).T

    def fit(
        self,
        dataframe: pd.DataFrame,
        column_names: types.ColumnNamesType,
    ):
        """
        Fit to calculate iqr parameters for the selected columns.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to calculate quantiles.
        column_names : ColumnNamesType
            One or number of columns in the dataframe.

        Returns
        -------
        self : object
            Instance object.
        """
        self.params["column_names"] = wrap_cols(column_names)
        self._check_cols(dataframe, self.params["column_names"])
        self.__calculate_params(dataframe)
        self.fitted = True
        return self

    def transform(self, dataframe: pd.DataFrame, inplace: bool = False) -> Union[pd.DataFrame, None]:
        """
        Remove objects from the dataframe which are behind maximum and minimum
        values of boxplots for each metric distribution.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to transform.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self._check_fitted()
        self._check_cols(dataframe, self.params["column_names"])
        if self.verbose:
            prev_stats: List[Dict[str, float]] = log.RobustLogger.get_stats(dataframe, self.params["column_names"])

        transformed: pd.DataFrame = dataframe if inplace else dataframe.copy()
        iqr: np.ndarray = self.params["quartiles"][:, 1] - self.params["quartiles"][:, 0]
        tail: np.ndarray = self.params["quartiles"][:, 0] - 1.5 * iqr
        head: np.ndarray = self.params["quartiles"][:, 1] + 1.5 * iqr
        mask: pd.Series = (
            (transformed[self.params["column_names"]] < tail) | (transformed[self.params["column_names"]] > head)
        ).any(axis=1)
        bad_ids = transformed.loc[mask].index
        transformed.drop(bad_ids, inplace=True)

        if self.verbose:
            log.info_log(f"Making IQR transformation of columns {self.params['column_names']}")
            new_stats: Dict[str, float] = log.RobustLogger.get_stats(transformed, self.params["column_names"])
            log.RobustLogger.verbose_list(prev_stats, new_stats, self.params["column_names"])
        return None if inplace else transformed

    def fit_transform(
        self,
        dataframe: pd.DataFrame,
        column_names: types.ColumnNamesType,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Fit preprocessor parameters using given dataframe and transform it.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to calculate quantiles and for further transformation.
        column_names : ColumnNamesType
            One or number of columns in the dataframe.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self.fit(dataframe, column_names)
        return self.transform(dataframe, inplace)



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
from joblib import Parallel, delayed, parallel_backend

import ambrosia.tools.pvalue_tools as pvalue_pkg
import ambrosia.tools.stat_criteria as criteria_pkg
from ambrosia import types
from ambrosia.tools.ab_abstract_component import ABStatCriterion
from ambrosia.tools.decorators import filter_kwargs

AVAILABLE_AB_CRITERIA: Dict[str, ABStatCriterion] = {
    "ttest": criteria_pkg.TtestIndCriterion,
    "ttest_rel": criteria_pkg.TtestRelCriterion,
    "mw": criteria_pkg.MannWhitneyCriterion,
    "wilcoxon": criteria_pkg.WilcoxonCriterion,
}


def inject_effect(
    sampled_metric_vals: np.ndarray,
    sample_size_a: int,
    effect: float,
    modeling_method: str = "constant",
    variation_factor: float = 10,
    random_seed: Optional[int] = None,
) -> np.ndarray:
    """
    Inject effect to values of group B.

    Parameters
    ----------
    sampled_metric_vals : np.ndarray
        Samples of groups A/B: |group A values|group B values|.
    sample_size_a : int
        Size of group A in ``sampled_metric_vals``, i.e.
        first ``sample_size_a`` elements correspond to group A.
    effect : float
        Effect for modifying group B,
        for example value = value * effect.
    modeling_method : str, default: ``constant``
        Method to modify group B
        ``constant``: value = value * effect
        ``shift``: value = value + value_mean * effect
        ``normal``: Add gauss noise Normal(mean(A) * (1 - effect),
                    Variation(A) / variation_factor).
    variation_factor : float, default: ``10``
        Factor for variation in ``normal`` method.

    Returns
    -------
    modified_metric_vals : np.ndarray
        Modified metric in group B and same values in group A.
    """
    available_modeling_mehods: List[str] = ["constant", "shift", "normal"]
    rng = np.random.default_rng(random_seed)
    modified_metric_vals: np.ndarray = sampled_metric_vals.copy()
    mean: np.ndarray = modified_metric_vals[sample_size_a:, :].mean()
    bs_size: int = modified_metric_vals.shape[1]
    sample_size_b: int = len(modified_metric_vals) - sample_size_a

    if modeling_method == "constant":
        modified_metric_vals[sample_size_a:, :] = effect * modified_metric_vals[sample_size_a:, :]
    elif modeling_method == "shift":
        modified_metric_vals[sample_size_a:, :] = modified_metric_vals[sample_size_a:, :] + (effect - 1) * mean
    elif modeling_method == "normal":
        effect_delta = (effect - 1) * mean
        effect_std = modified_metric_vals[sample_size_a:, :].std(ddof=1) / variation_factor
        modified_metric_vals[sample_size_a:, :] = modified_metric_vals[sample_size_a:, :] + rng.normal(
            loc=effect_delta, scale=effect_std, size=(sample_size_b, bs_size)
        )
    else:
        raise ValueError(
            f"Effect modeling method {modeling_method} is not found, chose from {available_modeling_mehods}"
        )
    return modified_metric_vals


def estim_stat_criterion_power(
    sampled_metric_vals: np.ndarray,
    sample_size_a: int,
    criterion: types.CompoundCriterionType = "ttest",
    alpha: float = 0.05,
    **kwargs,
) -> float:
    """
    Estimate power of statistical criterion.

    Parameters
    ----------
    sampled_metric_vals : np.ndarray
        Sampled metrics for groups A/B
        |group A values|group B values|
    sample_size_a : int
        Size of group A
    criterion : Union[Callable[[np.ndarray, np.ndarray], CriterionResultType], str], default: ``"ttest"``
        Statistical criterion - function f: f(group A, group B) = (statistic, p_value)
        or name of criterion as string, for example 'ttest'
    alpha : float, default: ``0.05``
        First type error bound, 1 - alpha: correctness
    **kwargs : Dict
        Keyword arguments for statistical criterion.

    Returns
    -------
    power : float
        Empirical bootstraped estimation for power
    """
    a_group_metrics: np.ndarray = sampled_metric_vals[:sample_size_a]
    b_group_metrics: np.ndarray = sampled_metric_vals[sample_size_a:]

    if isinstance(criterion, str) & (criterion in AVAILABLE_AB_CRITERIA):
        criterion = AVAILABLE_AB_CRITERIA[criterion]
    elif not (hasattr(criterion, "calculate_pvalue") and callable(criterion.calculate_pvalue)):
        raise ValueError(
            f"Choose correct criterion name from {list(AVAILABLE_AB_CRITERIA)} or pass correct custom class"
        )
    power: float = np.mean(criterion().calculate_pvalue(a_group_metrics, b_group_metrics, **kwargs) <= alpha)
    return power


def get_bs_stat(sample: np.ndarray, stat: str = "mean", N: int = 1000, random_seed: Optional[int] = None) -> np.ndarray:
    """
    Evaluate statistic (mean / median) using bootstrap method.

    Parameters
    ----------
    sample : np.ndarray
        Given sample array
    stat : str, default: ``"mean"``
        Name of statistic to be calculated
    N : int, default: ``1000``
       Bootstrap size
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.

    Returns
    -------
    bs_stat : np.ndarray
        Statistic calculated via bootstrap
    """
    rng = np.random.default_rng(random_seed)
    permissible_string_statistics: List[str] = ["mean", "median"]
    bs_samples: np.ndarray = rng.choice(sample, replace=True, size=(len(sample), N))
    if stat == "mean":
        bs_stat: np.ndarray = np.mean(bs_samples, axis=0)
    elif stat == "median":
        bs_stat: np.ndarray = np.median(bs_samples, axis=0)
    else:
        raise ValueError(f'Statistic is not found, choose from {", ".join(permissible_string_statistics)}')
    return bs_stat


def get_bs_sample_stat(
    sample: np.ndarray,
    sample_size_a: int,
    alpha: float,
    N: int = 1000,
    stat: str = "mean",
    random_seed: Optional[int] = None,
    alternative: str = "two-sided",
) -> bool:
    """
    Evaluate if difference in groups is significant.

    If confidence interval contains 0 then effect is not statistically significant.
    Returns ``True`` if we have effect else ``False``.

    Parameters
    ----------
    sample : np.ndarray
        Sample with groups A and B
    sample_size_a : int
        Size of group A => sample = |group A| group B|
    alpha : float
        Bound for first type error
    N : int, default: ``1000``
       Bootstrap size
    stat : str, default: ``"mean"``
          Name of calculated statistic, for example 'mean'
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    alternative : str, default: ``"two-sided"``
        Alternative hypothesis, can be ``"two-sided"``, ``"greater"``
        or ``"less"``.

    Returns
    -------
    overlap : bool
        True => H_0 is rejected <-> there is effect in groups
        False => H_0 is not rejected <-> there is no effect in groups
    """
    rng = np.random.SeedSequence(random_seed)
    seed_sequence: np.ndarray = rng.generate_state(2)
    bs_stat_a = get_bs_stat(sample[:sample_size_a], stat=stat, N=N, random_seed=seed_sequence[0])
    bs_stat_b = get_bs_stat(sample[sample_size_a:], stat=stat, N=N, random_seed=seed_sequence[1])
    bs_stat_diff: np.ndarray = bs_stat_b - bs_stat_a
    if alternative == "less":
        right_side = np.quantile(bs_stat_diff, 1 - alpha)
        overlap = right_side <= 0
    elif alternative == "greater":
        left_side = np.quantile(bs_stat_diff, alpha)
        overlap = left_side >= 0
    elif alternative == "two-sided":
        left_side, right_side = np.quantile(bs_stat_diff, [alpha / 2.0, 1 - alpha / 2.0])
        overlap = not left_side <= 0 <= right_side
    else:
        raise ValueError(f"Incorrect alternative value - {alternative}, choose from two-sided, less, greater")
    return overlap


def make_bootstrap(
    sampled_metric_vals: np.ndarray,
    sample_size_a: int,
    alpha: float = 0.05,
    N: int = 1000,
    stat: str = "mean",
    random_seed: Optional[int] = None,
    n_jobs: int = 1,
    backend: str = "loky",
    verbose: bool = False,
    **kwargs,
) -> float:
    """
    Evaluate share of cases when we find difference in groups using bootstrap.
    We can use parallel mode for faster evaluations in case of large data.

    Parameters
    ----------
    sampled_metric_vals : np.ndarray
         Metric values in both groups.
    sample_size_a : int
        Size of group A => |group A| group B|.
    alpha : float, default: ``0.05``
        Bound for the first type error.
    N : int, default: ``1000``
       Number of bootstraps.
    stat : str, default: ``"mean"``
        Statistics to be calculated.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    backend : str, default: ``"loky"``
        Type of backend for joblib parallel computation.
    verbose : bool, default: ``False``
        Whether to make logging.
    **kwargs : Dict
        Other keyword arguments.

    Returns
    -------
    test_result : float
        Empirical estimation for power.
    """
    num_samples = sampled_metric_vals.shape[1]
    rng = np.random.SeedSequence(random_seed)
    seed_sequence: np.ndarray = rng.generate_state(num_samples)
    iterator = zip(range(num_samples), seed_sequence)
    with parallel_backend(backend=backend, n_jobs=n_jobs):
        results_parallel = Parallel(verbose=verbose)(
            delayed(get_bs_sample_stat)(
                sample=sampled_metric_vals[:, sample_num],
                sample_size_a=sample_size_a,
                alpha=alpha,
                N=N,
                stat=stat,
                random_seed=seed,
                **kwargs,
            )
            for sample_num, seed in iterator
        )
    return np.mean(results_parallel)


def eval_error(
    sampled_metric_vals: np.ndarray,
    sample_size_a: int,
    alpha: float,
    mode: str = "ttest",
    stat: str = "mean",
    bootstrap_size: int = 1000,
    random_seed: Optional[int] = None,
    n_jobs: int = 1,
    verbose: bool = False,
    **kwargs,
) -> float:
    """
    Evaluate I type error/power of the experiment.

    Parameters
    ----------
    sampled_metric_vals : np.ndarray
        Samples of A/B groups: |group A values|group B values|.
    sample_size_a : int
        Size of  the group A in ``sampled_metric_vals``, i.e.
        first ``sample_size_a`` elements correspond to the group A.
    alpha : float
        First type error bound, 1 - alpha: correctness.
    mode : str, default: ``"ttest"``
        Statistical criterion, for example ``'ttest'``.
    stat : str, default: ``mean``
        Statistic to be calculated for sample groups during a bootstrap.
    bootstrap_size : int, default: ``1000``
        Number of bootstrap of A/B pairs.
    random_seed : int, optional
        A seed for the deterministic outcome of random processes.
    n_jobs : int, default: ``1``
        Amount of threads/workers for parallel.
    verbose : bool, default: ``False``
        Whether use logging.
    **kwargs : Dict
        Keyword arguments for statistical criterion.

    Returns
    -------
    error : float
        Second type error estimation or correctness, i.e.
        1 - P_{A=B} (criterion is completed) - correctness
        P_{A!=B} (criterion is completed) - second type error.
    """
    not_bootstrap_criteria: List[str] = ["ttest", "ttest_rel", "mw", "wilcoxon"]
    if mode in not_bootstrap_criteria:
        power: float = estim_stat_criterion_power(
            sampled_metric_vals, sample_size_a, criterion=mode, alpha=alpha, **kwargs
        )
    elif mode == "bootstrap":
        power: float = make_bootstrap(
            sampled_metric_vals,
            sample_size_a,
            alpha,
            N=bootstrap_size,
            stat=stat,
            random_seed=random_seed,
            n_jobs=n_jobs,
            verbose=verbose,
            **kwargs,
        )
    else:
        raise ValueError(f"Criterion {mode} is not found, choose from {not_bootstrap_criteria} or 'bootstrap'")
    return power


class BootstrapStats:
    """
    Generation empirical distribution for statistic(group A values, group B values).

    Attributes
    ----------
    All attributes are private.

    Methods
    -------
    fit(group_a: Iterable[float], group_b: Iterable[float])
        Fits the empirical distribution using values from group_a and group_b

    confidence_interval(confidence_level: Union[float, Iterable[float]]=0.95)
        Build confidence interval using empirical distribution from fit method with
        given confidence level

    pvalue_criterion()
        Calculate pvalue using confidence interval as criterion

    """

    def __init__(self, bootstrap_size: int = 10000, metric: Union[str, Callable] = "mean", paired: bool = False):
        """
        Parameters
        ----------
        bootstrap_size: int
            Amount of bootstrap groups
        metric: str or callable
            Metric to be calculated - mean or fraction
        paired: bool, default False
            If True use paired sampling, could be usefull for paired groups
        """
        self.__bs_size = bootstrap_size
        self.__metric_distribution = np.nan
        if isinstance(metric, str):
            accepted_str_metrics: List[str] = ["mean", "fraction", "median"]
            if metric not in accepted_str_metrics:
                raise ValueError(f'Choose metric name from - {", ".join(accepted_str_metrics)}')
        self.__metric = metric
        self.__min_of_distribution = None
        self.__max_of_distribution = None
        if isinstance(paired, bool):
            self.__paired = paired
        else:
            raise ValueError("Parameter paired can only take boolean values")

    def __handle_str_metric(self, bootstrap_a: np.ndarray, bootstrap_b: np.ndarray) -> None:
        """
        Handle case if metric is string.
        """
        if self.__metric == "mean":
            self.__metric_distribution = np.mean(bootstrap_b, axis=1) - np.mean(bootstrap_a, axis=1)
        elif self.__metric == "fraction":
            self.__metric_distribution = np.mean(bootstrap_b, axis=1) / np.mean(bootstrap_a, axis=1) - 1
        elif self.__metric == "median":
            self.__metric_distribution = np.median(bootstrap_b, axis=1) - np.median(bootstrap_a, axis=1)

    def __handle_std_value(self) -> float:
        """
        Calculate value for criterion.
        """
        if isinstance(self.__metric, str):
            if self.__metric in ["mean", "fraction", "median"]:
                val = 0
        else:
            val = self.__metric(np.array([1]), np.array([1]))
        return val

    def __handle_sampling(
        self, group_a: Iterable[float], group_b: Iterable[float], random_seed: Optional[int] = None
    ) -> Tuple[np.ndarray, np.ndarray]:
        rng = np.random.default_rng(random_seed)
        if self.__paired:
            a_size, b_size = len(group_a), len(group_b)
            if a_size != b_size:
                err: str = f"Paired groups must have equal sizes, have - {len(group_a)} and {len(group_b)}"
                raise ValueError(err)
            idxs: np.ndarray = rng.choice(np.arange(a_size), size=(self.__bs_size, a_size))
            return group_a[idxs], group_b[idxs]
        return (
            rng.choice(group_a, size=(self.__bs_size, len(group_a))),
            rng.choice(group_b, size=(self.__bs_size, len(group_b))),
        )

    @filter_kwargs
    def fit(self, group_a: Iterable[float], group_b: Iterable[float], random_seed: Optional[int] = None) -> None:
        """
        Make bootstrap samples from given groups.
        Calculates and store empiric distribution for saved metric in __init__

        Parameters
        ----------
        group_a : Iterable[float]
            Values of A group
        group_b : Iterable[float]
            Values of B group
        random_seed : int, optional
            A seed for the deterministic outcome of random bootstrap processes.

        Returns
        -------
            Nothing
        """
        group_a = np.array(group_a)
        group_b = np.array(group_b)
        bootstraped_a_group, bootstraped_b_group = self.__handle_sampling(group_a, group_b, random_seed)
        if isinstance(self.__metric, str):
            self.__handle_str_metric(bootstraped_a_group, bootstraped_b_group)
        else:
            self.__metric_distribution = self.__metric(bootstraped_a_group, bootstraped_b_group)

    def min_of_distrbution(self) -> float:
        """
        Minimum empirical distribution
        """
        if self.__min_of_distribution is None:
            self.__min_of_distribution = np.min(self.__metric_distribution)
        return self.__min_of_distribution

    def max_of_distribution(self) -> float:
        """
        Maximum empirical distribution
        """
        if self.__max_of_distribution is None:
            self.__max_of_distribution = np.max(self.__metric_distribution)
        return self.__max_of_distribution

    @filter_kwargs
    def confidence_interval(
        self, confidence_level: Union[float, Iterable[float]] = 0.95, alternative: str = "two-sided"
    ) -> types.IntervalType:
        """
        Returns bootstraped confidence interval, based on fit method.

        Parameters
        ----------
        confidence_level: Union[float, Iterable[float]], default: ``0.95``
            Bounds for error, that is
            Pr (mean(metric) not in interval) <= alpha
        alternative : str, default: ``"two-sided"``
                Alternative for static criteria - two-sided, less, greater
                Less means, that mean in first group less, than mean in second group
        Returns
        -------
        interval : IntervalType
            Confidence interval for each error in alpha
        """
        if not hasattr(self, "_BootstrapStats__metric_distribution"):
            raise AttributeError("Use method fit to build empirical distribution on metric")

        alpha: Union[float, Iterable[float]] = 1 - confidence_level
        alpha = pvalue_pkg.corrected_alpha(alpha, alternative)
        if not isinstance(alpha, float):
            alpha = np.array(alpha)
        left_bounds: np.ndarray = np.quantile(self.__metric_distribution, q=alpha / 2)
        right_bounds: np.ndarray = np.quantile(self.__metric_distribution, q=1 - alpha / 2)

        return pvalue_pkg.choose_from_bounds(
            left_bounds,
            right_bounds,
            alternative,
            left_bound=self.min_of_distrbution(),
            right_bound=self.max_of_distribution(),
        )

    @filter_kwargs
    def pvalue_criterion(self, alternative: str = "two-sided") -> float:
        """
        Calculate pvalue for bootstrap criterion.

        Returns
        --------
        pvalue : float
            Corresponding pvalue
        alternative : str, defaulte: ``"two-sided"``
                Alternative for static criteria - two-sided, less, greater
                Less means, that mean in first group less, than mean in second group
        """
        criterion_value: float = self.__handle_std_value()
        return pvalue_pkg.calculate_pvalue_by_interval(
            BootstrapStats.confidence_interval, criterion_value, self=self, alternative=alternative, precision=10e-15
        )



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import Iterable, List

import numpy as np
import pandas as pd
import scipy.stats as sps

import ambrosia.tools._lib._bin_ci_aide as helper_dir
import ambrosia.tools.pvalue_tools as pvalue_pkg
from ambrosia import types

from . import EFFECT_COL_NAME, FIRST_TYPE_ERROR_COL_NAME, GROUP_SIZE_COL_NAME, STAT_ERRORS_COL_NAME

RELATIVE_ABSOLUTE_DELTA_ERROR = ValueError("Choose relative or absolute delta, not both")
ROUND_DIGITS: int = 3
ROUND_DIGITS_TABLE: int = 3
ROUND_DIGITS_PERCENT: int = 1


class BinomTwoSampleCI(ABC):
    """
    Implementation of two-sample confidence interval.
    Supported the following types  interval_type:

    X ~ Bin(m, p1)
    Y ~ Bin(n, p2)

    Estimated Delta = p1 - p2
    For more information explore:
    http://stat.wharton.upenn.edu/~lbrown/Papers/2005c%20Confidence%20intervals%20for%
    20the%20two%20sample%20binomial%20distribution%20problem.pdf

    Methods
    -------

        Frequency:
        ----------

        Wald's CI - Using MLE estimation for proportions and
                asymptotic normality

        Yule's CI - Using estimation for Var(p1 - p2), in assumption
                (p1 = p2) and (m ~= n). Could be use for criterion
                H0: p1 = p2   vs   H1: p1 != p2
        Yule's modified CI - analogue for previous one for m != n
        Newcombe's CI - combination of CI for p1 and p2, building CI for difference
        Recentred CI - Using t quantiles

        Bayes:
        ------

        Jeffrey's CI - Using for prior distribution Beta(1/2, 1/2)
        Agresti's CI - Using for prior distribution Beta(1, 1)
        Conjugate Beta - Using for prior Beta(n_success, n_failure)

    """

    __PRECISION: float = 0.0001

    # This class just a container with all confidence intervals.
    # Thats why there is no instantiation.
    @abstractmethod
    def __init__(self):
        pass

    @staticmethod
    def __wald_ci(
        sample_a: int, sample_b: int, a_trials: int, b_trials: int, significance_level: float
    ) -> types.IntervalType:
        """
        Implementation of Wald confidence interval.
        """
        p_a_est = sample_a / a_trials
        p_b_est = sample_b / b_trials
        q_a_est = 1 - p_a_est
        q_b_est = 1 - p_b_est
        delta_estimation = p_b_est - p_a_est

        quantile = sps.norm.ppf(significance_level / 2)
        variation = p_a_est * q_a_est / a_trials + p_b_est * q_b_est / b_trials
        shift = -quantile * np.sqrt(variation)

        return delta_estimation - shift, delta_estimation + shift

    @staticmethod
    def __yule_ci(
        sample_a: int, sample_b: int, a_trials: int, b_trials: int, significance_level: float, modified: bool = False
    ) -> types.IntervalType:
        """
        Implementation of Yule's confidence intervals.
        """

        p_a_est = sample_a / a_trials
        p_b_est = sample_b / b_trials
        delta_estimation = p_b_est - p_a_est

        quantile = sps.norm.ppf(significance_level / 2)

        if modified:
            coef_a = b_trials / a_trials
            coef_b = a_trials / b_trials
            p_hat = (sample_a * coef_a + sample_b * coef_b) / (a_trials + b_trials)
            q_hat = 1 - p_hat
            variation = (1 / a_trials + 1 / b_trials) * p_hat * q_hat
        else:
            p_overline = (sample_a + sample_b) / (a_trials + b_trials)
            q_overline = 1 - p_overline
            variation = (1 / a_trials + 1 / b_trials) * p_overline * q_overline

        shift = -quantile * np.sqrt(variation)
        return delta_estimation - shift, delta_estimation + shift

    @staticmethod
    def __bayes_conjugate_beta(
        sample_a: int,
        sample_b: int,
        a_trials: int,
        b_trials: int,
        n_success: int,
        n_failure: int,
        significance_level: float,
    ) -> types.IntervalType:
        """
        Implementation of bayes confidence intervals with cojugate distribution
        Beta(n_success, n_falure).
        """

        p_tilde_a = (sample_a + n_success) / (n_success + n_failure + a_trials)
        p_tilde_b = (sample_b + n_success) / (n_success + n_failure + b_trials)

        q_tilde_a = 1 - p_tilde_a
        q_tilde_b = 1 - p_tilde_b
        delta_tilde = p_tilde_b - p_tilde_a

        quantile = sps.norm.ppf(significance_level / 2)
        variation = p_tilde_a * q_tilde_a / a_trials + p_tilde_b * q_tilde_b / b_trials
        shift = -quantile * np.sqrt(variation)
        return delta_tilde - shift, delta_tilde + shift

    @staticmethod
    def __square_eq_newcombe(p_est: float, m: int, quantile: float) -> types.IntervalType:  # pylint: disable=C0103
        """
        Helper function for newcombe_ci.
        """

        coef_a = 1 + quantile**2 / m
        coef_b = -(2 * p_est + quantile**2 / m)
        coef_c = p_est**2
        d = coef_b**2 - 4 * coef_a * coef_c  # pylint: disable=C0103
        left = (-coef_b - np.sqrt(d)) / (2 * coef_a)
        right = (-coef_b + np.sqrt(d)) / (2 * coef_a)

        return left, right

    @staticmethod
    def __newcombe_ci(
        sample_a: int,
        sample_b: int,
        a_trials: int,
        b_trials: int,
        significance_level: float = 0.05,
    ) -> types.IntervalType:
        """
        Implementation of Newcombe's confidence intervals.
        """
        p_a_est = sample_a / a_trials
        p_b_est = sample_b / b_trials
        delta_est = p_b_est - p_a_est

        quantile = sps.norm.ppf(significance_level / 2)

        # Calculate intevalls for A and B group
        left_bound_group_a, right_bound_group_a = BinomTwoSampleCI.__square_eq_newcombe(p_a_est, a_trials, quantile)
        left_bound_group_b, right_bound_group_b = BinomTwoSampleCI.__square_eq_newcombe(p_b_est, b_trials, quantile)

        variance_a_left = left_bound_group_a * (1 - left_bound_group_a) / a_trials
        variance_a_right = right_bound_group_a * (1 - right_bound_group_a) / a_trials
        variance_b_left = left_bound_group_b * (1 - left_bound_group_b) / b_trials
        variance_b_right = right_bound_group_b * (1 - right_bound_group_b) / b_trials
        variance_left = variance_a_left + variance_b_right
        variance_right = variance_b_left + variance_a_right
        left_bound = delta_est + quantile * np.sqrt(variance_left)
        right_bound = delta_est - quantile * np.sqrt(variance_right)

        return left_bound, right_bound

    @staticmethod
    def __recentered_ci(
        sample_a: int, sample_b: int, a_trials: int, b_trials: int, significance_level: float
    ) -> types.IntervalType:
        """
        Implementation of recentered confidence intervals.
        """
        p_a_est = sample_a / a_trials
        p_b_est = sample_b / b_trials
        delta_estimation = p_b_est - p_a_est
        p_est = (b_trials * p_a_est + a_trials * p_b_est) / (a_trials + b_trials)
        q_est = 1 - p_est
        quantile = -sps.t.ppf(significance_level / 2, df=a_trials + b_trials - 2)
        coef = 1 + quantile**2 / (a_trials + b_trials)
        center = delta_estimation / coef
        var_est = (1 / a_trials + 1 / b_trials) * p_est * q_est
        shift = quantile * np.sqrt(coef * var_est - delta_estimation**2 / (a_trials + b_trials)) / coef
        return center - shift, center + shift

    @staticmethod
    def calculate_pvalue(
        a_success: int,
        b_success: int,
        a_trials: int,
        b_trials: int,
        interval_type: str = "wald",
        alternative: str = "two-sided",
        n_success_conjugate: int = None,
        n_failure_conjugate: int = None,
    ) -> float:
        """
        Calculate pvalue for confidence interval.
        pvalue(x) = inf_a {a | x \\in S_a }

        Parameters
        ----------
        a_success : int
            Samples from Bin(a_trials, p_a)
        b_success : int
            Sample from Bin(b_trials, p_b)
        a_trials : int
            Parameter for trials amount for group A
        b_trials : int
            Parameter for trials amount for group B
        confidence_level : float
            Pr ( [p1 - p2] in CI ) -> confidence_level, n -> infty
        interval_type : str, default : ``"wald"``
            One from [wald, yule, yule_modif, newcombe, jeffrey, agresti, bayes_beta, recenter]
        alternative : str, default : ``"two-sided"``
            Alternative for static criteria - two-sided, less, greater
            Less means, that mean in first group less, than mean in second group
        n_success : int
            Arguments for conjugate distribution Beta(n_success, n_failure) if interval_type = "bayes_beta"
        n_failure : int
            Arguments for conjugate distribution Beta(n_success, n_failure) if interval_type = "bayes_beta"

        Comment
        -------
        You can pass numpy arrays with same shapes for a_success, b_success and e.t.c

        Returns
        -------
        pvalue : float
            P-value of the interval-induced criterion
        """

        return pvalue_pkg.calculate_pvalue_by_interval(
            BinomTwoSampleCI.confidence_interval,
            0,
            a_success=a_success,
            a_trials=a_trials,
            b_success=b_success,
            b_trials=b_trials,
            n_success_conjugate=n_success_conjugate,
            n_failure_conjugate=n_failure_conjugate,
            interval_type=interval_type,
            alternative=alternative,
        )

    @staticmethod
    def confidence_interval(
        a_success: int,
        b_success: int,
        a_trials: int,
        b_trials: int,
        confidence_level: float,
        interval_type: str = "wald",
        alternative: str = "two-sided",
        n_success_conjugate: int = None,
        n_failure_conjugate: int = None,
    ) -> types.IntervalType:
        """
        Main function building confidence interval.

        Parameters
        ----------
        a_success : int
            Samples from Bin(a_trials, p_a)
        b_success : int
            Sample from Bin(b_trials, p_b)
        a_trials : int
            Parameter for trials amount for group A
        b_trials : int
            Parameter for trials amount for group B
        confidence_level : float
            Pr ( [p1 - p2] in CI ) -> confidence_level, n -> infty
        interval_type : str, default : ``"wald"``
            One from [wald, yule, yule_modif, newcombe, jeffrey, agresti, bayes_beta, recenter]
        alternative : str, default : ``"two-sided"``
            Alternative for static criteria - two-sided, less, greater
            Less means, that mean in first group less, than mean in second group
        n_success : int
            Arguments for conjugate distribution Beta(n_success, n_failure) if interval_type = "bayes_beta"
        n_failure : int
            Arguments for conjugate distribution Beta(n_success, n_failure) if interval_type = "bayes_beta"

        Note
        ----
        You can pass numpy arrays with same shapes for a_success, b_success and e.t.c

        Returns
        -------
        interval : Tuple[float, float]
        """

        valid_types: List[str] = [
            "wald",
            "yule",
            "yule_modif",
            "newcombe",
            "jeffrey",
            "agresti",
            "bayes_beta",
            "recenter",
        ]
        pvalue_pkg.check_alternative(alternative)
        significance_level = pvalue_pkg.corrected_alpha(1 - confidence_level, alternative)
        if interval_type == "wald":
            left_ci, right_ci = BinomTwoSampleCI.__wald_ci(a_success, b_success, a_trials, b_trials, significance_level)
        elif interval_type == "yule":
            left_ci, right_ci = BinomTwoSampleCI.__yule_ci(
                a_success, b_success, a_trials, b_trials, significance_level, modified=False
            )
        elif interval_type == "yule_modif":
            left_ci, right_ci = BinomTwoSampleCI.__yule_ci(
                a_success, b_success, a_trials, b_trials, significance_level, modified=True
            )
        elif interval_type == "newcombe":
            left_ci, right_ci = BinomTwoSampleCI.__newcombe_ci(
                a_success, b_success, a_trials, b_trials, significance_level
            )
        elif interval_type == "recenter":
            left_ci, right_ci = BinomTwoSampleCI.__recentered_ci(
                a_success, b_success, a_trials, b_trials, significance_level
            )
        elif interval_type == "jeffrey":
            left_ci, right_ci = BinomTwoSampleCI.__bayes_conjugate_beta(
                a_success, b_success, a_trials, b_trials, 0.5, 0.5, significance_level
            )
        elif interval_type == "agresti":
            left_ci, right_ci = BinomTwoSampleCI.__bayes_conjugate_beta(
                a_success, b_success, a_trials, b_trials, 1, 1, significance_level
            )
        elif interval_type == "bayes_beta":
            error_beta_params: str = "Pass correct params n_success_conjugate, n_faliure_conjugate"
            args_correctness = (
                (n_success_conjugate is not None)
                and (n_failure_conjugate is not None)
                and (n_success_conjugate > 0)
                and (n_failure_conjugate > 0)
            )
            if not args_correctness:
                raise ValueError(error_beta_params)
            left_ci, right_ci = BinomTwoSampleCI.__bayes_conjugate_beta(
                a_success, b_success, a_trials, b_trials, n_success_conjugate, n_failure_conjugate, significance_level
            )
        else:
            algo_type_error: str = f'Choose one from accepted methods, from - {", ".join(valid_types)}'
            raise ValueError(algo_type_error)
        left_ci, right_ci = pvalue_pkg.choose_from_bounds(
            left_ci, right_ci, alternative, left_bound=np.array([-1]), right_bound=np.array([1])
        )
        return left_ci, right_ci


def get_table_power_on_size_and_conversions(
    interval_type: str = "wald",
    p_a_values: Iterable[float] = (0.5,),
    p_b_values: Iterable[float] = (0.4,),
    sample_sizes: Iterable[int] = (100,),
    amount: int = 10000,
    confidence_level: float = 0.95,
) -> pd.DataFrame:
    """
    Table with power / empirical 1 type error = 1 - coverage, for fixed size and conversions.

    Parameters
    ----------
    interval_type : str, default : ``"wald"``
        interval_type for confidence interval
    p_a_values : Iterable[float], default : ``(0.5,)``
        Conversions for A group
    p_b_values : Itrable[float], default : ``(0.5,)``
        Conversions for B group
    sample_sizes : Iterable[float], default : ``(100,)``
        Sizes for samples
    amount : int, default : ``10000``
        Amount of generated samples for one n(trials amount), to estimate power
    confidence_level : float, default : ``0.95``
        Such value x, that: Pr ( delta in I ) >= x

    Returns
    -------
    table : pd.DataFrame
        Required table with power
    """
    trials = np.array(sample_sizes)
    conversions_cond = np.all((np.array(p_a_values) >= 0) & (np.array(p_a_values) <= 1)) and np.all(
        (np.array(p_b_values) >= 0) & (np.array(p_b_values) <= 1)
    )
    if not conversions_cond:
        raise ValueError("Conversions must be from 0 to 1")
    powers_array: List[np.ndarray] = []
    for p_a in p_a_values:
        for p_b in p_b_values:
            sample_a = sps.binom.rvs(n=trials, p=p_a, size=(amount, len(sample_sizes)))
            sample_b = sps.binom.rvs(n=trials, p=p_b, size=(amount, len(sample_sizes)))
            binom_kwargs = {
                "interval_type": interval_type,
                "a_success": sample_a,
                "b_success": sample_b,
                "a_trials": trials,
                "b_trials": trials,
                "confidence_level": confidence_level,
            }
            conf_interval: types.ManyIntervalType = BinomTwoSampleCI.confidence_interval(**binom_kwargs)
            power: np.ndarray = helper_dir.__helper_calc_empirical_power(conf_interval)
            powers_array.append(power)
    power_matrix = np.vstack(powers_array)
    table = pd.DataFrame(
        power_matrix,
        index=pd.MultiIndex.from_tuples(
            [(round(p_a, 3), round(p_b, 3)) for p_a in p_a_values for p_b in p_b_values], names=[r"$p_a$", r"$p_b$"]
        ),
        columns=sample_sizes,
    )
    table.index.name = "conversions"
    table.columns.name = "sample sizes"
    return table


def get_table_power_on_size_and_delta(
    p_a: float,
    sample_sizes: Iterable[int],
    first_errors: Iterable[float] = (0.05,),
    delta_values: Iterable[float] = None,
    delta_relative_values: Iterable[float] = None,
    interval_type: str = "wald",
    # alternative: str = "two-sided",
    amount: int = 10000,
    as_numeric: bool = False,
) -> pd.DataFrame:
    """
    Table with power / empirical 1 type error = 1 - coverage for fixed size and effect.

    Parameters
    ----------
    p_a : Iterable[float]
        Conversion in A group
    sample_sizes : Iterable[float]
        Sizes for samples
    first_errors : Iterable[float], default : ``(0.05,)``
        First type error values
    delta_values : Iterable[float]
        Absolute delta values: p_b - p_a = delta
    delta_relative_values : Iterable[float]
        Relative delta values: p_b = delta_relative * p_a
    interval_type : str
        interval_type for confidence interval
    alternative : str, default : ``"two-sided"``
        Alternative for static criteria - two-sided, less, greater
        Less means, that mean in first group less, than mean in second group
    amount : int, default : ``10000``
        Amount of generated samples for one n(trials amount), to estimate power
    as_numeric : bool, default: ``False``
        The result of calculations can be obtained as a percentage string
        either as a number, this parameter could used to toggle.
        Works only for relative values of power.

    Returns
    -------
    table : pd.DataFrame
        Required table with power
    """
    trials = np.array(sample_sizes)
    if not (delta_values is None) ^ (delta_relative_values is None):
        raise RELATIVE_ABSOLUTE_DELTA_ERROR
    if delta_values is not None:
        p_b_values: np.ndarray = p_a + np.array(delta_values)
        grid_delta: np.ndarray = delta_values
    else:
        p_b_values: np.ndarray = p_a * np.array(delta_relative_values)
        grid_delta: np.ndarray = [f"{np.round((x - 1) * 100, ROUND_DIGITS_PERCENT)}%" for x in delta_relative_values]
    if not np.all((p_b_values >= 0) & (p_b_values <= 1)):
        raise ValueError(f"Probability of success in group B must be positive, not {p_b_values}")

    values = [(round(a, ROUND_DIGITS), b) for a in first_errors for b in grid_delta]
    table: pd.DataFrame = pd.DataFrame(
        index=pd.MultiIndex.from_tuples(values, names=[FIRST_TYPE_ERROR_COL_NAME, EFFECT_COL_NAME]),
        columns=trials,
    )
    table.columns.name = GROUP_SIZE_COL_NAME
    sample_a = sps.binom.rvs(n=trials, p=p_a, size=(amount, trials.shape[0]))
    for alpha in first_errors:
        for p_b, delta in zip(p_b_values, grid_delta):
            sample_b = sps.binom.rvs(n=trials, p=p_b, size=(amount, trials.shape[0]))
            binom_kwargs = {
                "interval_type": interval_type,
                "a_success": sample_a,
                "b_success": sample_b,
                "a_trials": trials,
                "b_trials": trials,
                "confidence_level": 1 - alpha,
                #    "alternative": alternative,
            }
            conf_interval: types.ManyIntervalType = BinomTwoSampleCI.confidence_interval(**binom_kwargs)
            power: np.ndarray = helper_dir.__helper_calc_empirical_power(conf_interval)
            if as_numeric:
                power = [round(power_val, ROUND_DIGITS_TABLE) for power_val in power]
            else:
                power = [str(round(power_val * 100, ROUND_DIGITS_PERCENT)) + "%" for power_val in power]
            table.loc[(alpha, delta), trials] = power
    return table


def iterate_for_sample_size(
    interval_type: str,
    first_errors: Iterable[float],
    second_errors: Iterable[float],
    p_a: float,
    p_b_values: Iterable[float],
    grid_delta: Iterable[float],
    amount: int,
) -> pd.DataFrame:
    """
    Iterate over params for different sample size
    """
    # values = [(round(a, ROUND_DIGITS), round(b, ROUND_DIGITS)) for a in first_errors for b in second_errors]
    # table: pd.DataFrame = pd.DataFrame(
    #     index=pd.MultiIndex.from_tuples(values, names=[r"$\alpha$", r"$\beta$"]),
    #     columns=grid_delta,
    # )
    multiindex = pd.MultiIndex.from_tuples([(eff,) for eff in grid_delta], names=[EFFECT_COL_NAME])
    multicols = pd.MultiIndex.from_tuples(
        [(f"({alpha}; {beta})",) for alpha in first_errors for beta in second_errors],
        names=[STAT_ERRORS_COL_NAME],
    )
    table: pd.DataFrame = pd.DataFrame(index=multiindex, columns=multicols)
    for alpha in first_errors:
        for beta in second_errors:
            power = 1 - beta
            for p_b, delta in zip(p_b_values, grid_delta):
                trials = helper_dir.__helper_bin_search_for_size(
                    interval_type=interval_type,
                    confidence_level=1 - alpha,
                    p_a=p_a,
                    p_b=p_b,
                    amount=amount,
                    power=power,
                )
                table.loc[delta, f"({alpha}; {beta})"] = trials
    return table


def get_table_sample_size_on_effect(
    interval_type: str = "wald",
    first_errors: Iterable[float] = (0.05,),
    second_errors: Iterable[float] = (0.2,),
    p_a: float = 0.5,
    delta_values: Iterable[float] = None,
    delta_relative_values: Iterable[float] = None,
    amount: int = 10000,
) -> pd.DataFrame:
    """
    Table for sample sizes with given effect and errors.

    Parameters
    ----------
    interval_type : str, default : ``"wald"``
        interval_type for confidence interval
    first_errors : Iterable[float], default : ``(0.05,)``
        First type error values
    second_errors : Iterable[float], default : ``(0.2,)``
        Second type error values
    p_a : Iterable[float]
        Conversion in A group, default : ``0.5``
    delta_values : Iterable[float]
        Absolute delta values: p_b - p_a = delta
    delta_relative_values : Iterable[float]
        Relative delta values: p_b = delta_relative * p_a
    amount : int, default : ``1000``
        Amount of generated samples for one n(trials amount), to estimate power

    Returns
    -------
    table : pd.DataFrame
        Required table with sample sizes
    """

    errors_condition = np.all((np.array(first_errors) >= 0) & (np.array(first_errors) <= 1)) and np.all(
        (np.array(second_errors) >= 0) & (np.array(second_errors) <= 1)
    )
    if not errors_condition:
        raise ValueError("Errors must be from 0 to 1")

    if (delta_values is not None) and (delta_relative_values is not None):
        raise RELATIVE_ABSOLUTE_DELTA_ERROR

    # If delta type not set => set to absolute
    if delta_values is None and delta_relative_values is None:
        delta_values = [p_a / 2]

    if delta_values is not None:
        p_b_values: np.ndarray = p_a + np.array(delta_values)
        grid_delta: np.ndarray = delta_values
    else:
        p_b_values: np.ndarray = p_a * np.array(delta_relative_values)
        grid_delta: np.ndarray = [f"{np.round((x - 1) * 100, ROUND_DIGITS_PERCENT)}%" for x in delta_relative_values]
    if not np.all((p_b_values >= 0) & (p_b_values <= 1)):
        raise ValueError(f"Probability of success in group B must be positive, not {p_b_values}")
    table = iterate_for_sample_size(interval_type, first_errors, second_errors, p_a, p_b_values, grid_delta, amount)
    return table


def iterate_for_delta(
    interval_type: str,
    first_errors: Iterable[float],
    second_errors: Iterable[float],
    sample_sizes: Iterable[int],
    p_a: float,
    amount: int,
    delta_type: str,
    as_numeric: bool = False,
) -> pd.DataFrame:
    """
    Helps to find effect for different params.
    """
    multiindex = pd.MultiIndex.from_tuples([(trials,) for trials in sample_sizes], names=[GROUP_SIZE_COL_NAME])
    multicols = pd.MultiIndex.from_tuples(
        [(f"({alpha}; {beta})",) for alpha in first_errors for beta in second_errors],
        names=[STAT_ERRORS_COL_NAME],
    )
    table: pd.DataFrame = pd.DataFrame(index=multiindex, columns=multicols)
    for alpha in first_errors:
        for beta in second_errors:
            power = 1 - beta
            for trials in sample_sizes:
                delta = helper_dir.__helper_bin_search_for_delta(
                    interval_type=interval_type,
                    confidence_level=1 - alpha,
                    p_a=p_a,
                    trials=trials,
                    amount=amount,
                    power=power,
                )
                if delta is not None and delta_type == "relative":
                    if as_numeric:
                        delta = round(abs(delta) / p_a, ROUND_DIGITS_TABLE) + 1
                    else:
                        delta = str(round(abs(delta) / p_a * 100, ROUND_DIGITS_PERCENT)) + "%"
                table.loc[trials, f"({alpha}; {beta})"] = delta
    return table


def get_table_effect_on_sample_size(
    interval_type: str = "wald",
    first_errors: Iterable[float] = (0.05,),
    second_errors: Iterable[float] = (0.2,),
    sample_sizes: Iterable[int] = (100,),
    p_a: float = 0.5,
    amount: int = 10000,
    delta_type: str = "relative",
    as_numeric: bool = False,
) -> pd.DataFrame:
    """
    Table for effects with given sample sizes and erros.
    If there are no effects satisfy first and second errors value will be set to None

    Parameters
    ----------
    interval_type : str, default : ``"wald"``
        interval_type for confidence interval
    first_errors : Iterable[float], default : ``(0.05,)``
        First type error values
    second_errors : Iterable[float], default : ``(0.2,)``
        Second type error values
    p_a : Iterable[float], default : ``0.5``
        Conversion in A group
    amount : int, default : ``10000``
        Amount of generated samples for one n(trials amount), to estimate power
    sample_sizes : Iterable[int], default : ``(100,)``
        Sample sizes for A/B group
    delta_type : str, default : ``"absolute``
        absolute or relative, if relative gives effect in percents: |delta| / p_a
    as_numeric : bool, default: ``False``
        The result of calculations can be obtained as a percentage string
        either as a number, this parameter could used to toggle.
        Works only for relative values of effect.

    Returns
    -------
    table : pd.DataFrame
        Required table with effects
    """
    errors_condition = np.all((np.array(first_errors) >= 0) & (np.array(first_errors) <= 1)) and np.all(
        (np.array(second_errors) >= 0) & (np.array(second_errors) <= 1)
    )
    if not errors_condition:
        error_mesage_errors: str = "Errors must be from 0 to 1"
        raise ValueError(error_mesage_errors)

    delta_types: List[str] = ["absolute", "relative"]
    if delta_type not in delta_types:
        raise ValueError(f"Delta type must be absolute relative, not {delta_type}")
    table: pd.DataFrame = iterate_for_delta(
        interval_type, first_errors, second_errors, sample_sizes, p_a, amount, delta_type, as_numeric
    )
    return table



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Handlers for different dataframe types.

Module containes functions and classes that help to deal
with data of different type during the experiment design problem.

These objects are used in `Designer` core class.
"""
import warnings
from typing import List

import pandas as pd

import ambrosia.spark_tools.empiric as empiric_spark
import ambrosia.spark_tools.theory as theory_spark
import ambrosia.tools.theoretical_tools as theory_pkg
import ambrosia.tools.tools as empiric_pkg
from ambrosia import types
from ambrosia.tools.ab_abstract_component import SimpleDesigner
from ambrosia.tools.import_tools import spark_installed

if spark_installed():
    import pyspark.sql.functions as spark_functions


DATA: str = "dataframe"
AVAILABLE: List[str] = ["pandas", "spark"]
AVAILABLE_TABLES_ERROR = TypeError(f'Type of table must be one of {", ".join(AVAILABLE)}')


class TheoryHandler(SimpleDesigner):
    """
    Unit for theory design.
    """

    def size_design(self, **kwargs) -> pd.DataFrame:
        return self._handle_cases(theory_pkg.design_groups_size, theory_spark.design_groups_size, **kwargs)

    def effect_design(self, **kwargs) -> pd.DataFrame:
        return self._handle_cases(theory_pkg.design_effect, theory_spark.design_effect, **kwargs)

    def power_design(self, **kwargs) -> pd.DataFrame:
        return self._handle_cases(theory_pkg.design_power, theory_spark.design_power, **kwargs)


class EmpiricHandler(SimpleDesigner):
    """
    Unit for empiric design.
    """

    def size_design(self, **kwargs) -> pd.DataFrame:
        return self._handle_cases(empiric_pkg.get_empirical_table_sample_size, empiric_spark.get_table_size, **kwargs)

    def effect_design(self, **kwargs) -> pd.DataFrame:
        return self._handle_cases(empiric_pkg.get_empirical_mde_table, empiric_spark.get_table_effect, **kwargs)

    def power_design(self, **kwargs) -> pd.DataFrame:
        if isinstance(kwargs[DATA], types.SparkDataFrame):
            kwargs["group_sizes"] = kwargs["sample_sizes_a"]
            del kwargs["sample_sizes_a"]
            del kwargs["sample_sizes_b"]
        return self._handle_cases(empiric_pkg.get_empirical_table_power, empiric_spark.get_table_power, **kwargs)


def calc_prob_control_class(table: types.PassedDataType, metric: types.MetricNameType) -> float:
    """
    Calculate conversion on binary metric for pandas or Spark dataframe.

    Parameters
    ----------
    table : SparkDataFrame or pd.DataFrame
        Table with binary metric.
    metric : MetricNameType
        Table Column name that containes binary metric of interest.

    Returns
    -------
    p_a : float
        Conversion in control group.
    """
    warning_message_values: str = "Metric values are not binary, choose empiric or theory method!"
    if isinstance(table, pd.DataFrame):
        if not set(table[metric].unique()).issubset({0, 1}):
            warnings.warn(warning_message_values)
        p_a = table[metric].mean()
    else:
        if not set(table.select(metric).distinct().toPandas()[metric]).issubset({0, 1}):
            warnings.warn(warning_message_values)
        p_a = table.select(spark_functions.mean(metric)).collect()[0][0]
    return p_a
